@@ -1,19 +1,23 @@
-"""Package init."""
+from public import public as _public
 
 from flufl.lock._lockfile import (
-    AlreadyLockedError, Lock, LockError, NotLockedError, SEP, TimeOutError)
-from public import public
+    AlreadyLockedError, Lock, LockError, LockState, NotLockedError, SEP,
+    TimeOutError)
 
 
-__version__ = '3.2'
+__version__ = '5.0.1'
 
 
-public(
+_public(
     AlreadyLockedError=AlreadyLockedError,
     Lock=Lock,
     LockError=LockError,
+    LockState=LockState,
     NotLockedError=NotLockedError,
     SEP=SEP,
     TimeOutError=TimeOutError,
     __version__=__version__,
     )
+
+
+del _public
