@@ -1,34 +1,3 @@
-"""Portable, NFS-safe file locking with timeouts for POSIX systems.
-
-This code implements an NFS-safe file-based locking algorithm influenced by
-the GNU/Linux open(2) manpage, under the description of the O_EXCL option:
-
-    [...] O_EXCL is broken on NFS file systems, programs which rely on it
-    for performing locking tasks will contain a race condition.  The
-    solution for performing atomic file locking using a lockfile is to
-    create a unique file on the same fs (e.g., incorporating hostname and
-    pid), use link(2) to make a link to the lockfile.  If link() returns
-    0, the lock is successful.  Otherwise, use stat(2) on the unique file
-    to check if its link count has increased to 2, in which case the lock
-    is also successful.
-
-The assumption made here is that there will be no 'outside interference',
-e.g. no agent external to this code will ever link() to the specific lock
-files used.
-
-Lock objects support lock-breaking so that you can't wedge a process forever.
-This is especially helpful in a web environment, but may not be appropriate
-for all applications.
-
-Locks have a 'lifetime', which is the maximum length of time the process
-expects to retain the lock.  It is important to pick a good number here
-because other processes will not break an existing lock until the expected
-lifetime has expired.  Too long and other processes will hang; too short and
-you'll end up trampling on existing process locks -- and possibly corrupting
-data.  In a distributed (NFS) environment, you also need to make sure that
-your clocks are properly synchronized.
-"""
-
 import os
 import sys
 import time
@@ -36,15 +5,33 @@ import errno
 import random
 import socket
 import logging
-import datetime
 
+from contextlib import suppress
+from datetime import datetime, timedelta
+from enum import Enum
 from logging import NullHandler
+from types import TracebackType
+from typing import cast, List, Optional, Tuple, Type, Union
+
+from psutil import pid_exists
 from public import public
 
 
-DEFAULT_LOCK_LIFETIME = datetime.timedelta(seconds=15)
+try:
+    from typing import Literal
+except ImportError:                                 # pragma: nocover
+    from typing_extensions import Literal  # type: ignore
+
+
+Interval = Union[timedelta, int]
+
+# 2020-06-26(bwarsaw): Once Python 3.8 is the minimum, we could annotatey some
+# of these module globals as typing.Final.
+
+
+DEFAULT_LOCK_LIFETIME = timedelta(seconds=15)
 # Allowable a bit of clock skew.
-CLOCK_SLOP = datetime.timedelta(seconds=10)
+CLOCK_SLOP = timedelta(seconds=10)
 MAXINT = sys.maxsize
 
 # Details separator; also used in calculating the claim file path.  Lock files
@@ -85,133 +72,251 @@ class TimeOutError(LockError):
     """The timeout interval elapsed before the lock succeeded."""
 
 
+@public
+class LockState(Enum):
+    """Infer the state of the lock.
+
+    There are cases in which it is impossible to infer the state of the lock,
+    due to the distributed nature of the locking mechanism and environment.
+    However it is possible to provide some insights into the state of the
+    lock.  Note that the policy on what to do with this information is left
+    entirely to the user of the library.
+    """
+    #: There is no lock file so the lock is unlocked.
+    unlocked = 1
+    #: We own the lock and it is fresh.
+    ours = 2
+    #: We own the lock but it has expired.  Another process trying
+    #: to acquire the lock will break it.
+    ours_expired = 3
+    #: We don't own the lock; the hostname in the details matches our
+    #: hostname and there is no pid running that matches pid.  Therefore,
+    #: the lock is stale.
+    stale = 4
+    #: Some other process owns the lock (probably) but it has expired.  Another
+    #: process trying to acquire the lock will break it.
+    theirs_expired = 5
+    #: We don't own the lock; either our hostname does not match the
+    #: details, or there is a process (that's not us) with a matching pid.
+    #: The state of the lock is therefore unknown.
+    unknown = 6
+
+
 @public
 class Lock:
-    """A portable way to lock resources by way of the file system."""
-    def __init__(self, lockfile, lifetime=None, separator=SEP):
-        """Create the resource lock using the given file name and lifetime.
+    """Portable, NFS-safe file locking with timeouts for POSIX systems.
 
-        Each process laying claim to this resource lock will create their own
-        temporary lock file based on the path specified.  An optional lifetime
-        is the length of time that the process expects to hold the lock.
+    This class implements an NFS-safe file-based locking algorithm
+    influenced by the GNU/Linux open(2) manpage, under the description
+    of the O_EXCL option:
 
-        :param lockfile: The full path to the lock file.
-        :param lifetime: The expected maximum lifetime of the lock, as a
-            timedelta.  Defaults to 15 seconds.
-        :param separator: The separator character to use in the lock file's
-            file name.  Defaults to the vertical bar (`|`) on POSIX systems
-            and caret (`^`) on Windows.
+        [...] O_EXCL is broken on NFS file systems, programs which rely on it
+        for performing locking tasks will contain a race condition.  The
+        solution for performing atomic file locking using a lockfile is to
+        create a unique file on the same fs (e.g., incorporating hostname and
+        pid), use link(2) to make a link to the lockfile.  If link() returns
+        0, the lock is successful.  Otherwise, use stat(2) on the unique file
+        to check if its link count has increased to 2, in which case the lock
+        is also successful.
+
+    The assumption made here is that there will be no *outside interference*,
+    e.g. no agent external to this code will ever link() to the specific lock
+    files used.
+
+    The user specifies a *lock file* in the constructor of this class.  This
+    is whatever file system path the user wants to coordinate locks on.  When
+    a process attempts to acquire a lock, it first writes a *claim file* which
+    contains unique details about the lock being acquired (e.g. the lock file
+    name, the hostname, the pid of the process, and a random integer).  Then
+    it attempts to create a hard link from the claim file to the lock file.
+    If no other process has the lock, this hard link succeeds and the process
+    accquires the lock.  If another process already has the lock, the hard
+    link will fail and the lock will not be acquired.  What happens in this
+    and other error conditions are described in the more detailed
+    documentation.
+
+    Lock objects support lock-breaking so that you can't wedge a process
+    forever.  This is especially helpful in a web environment, but may
+    not be appropriate for all applications.
+
+    Locks have a ``lifetime``, which is the maximum length of time the
+    process expects to retain the lock.  It is important to pick a good
+    number here because other processes will not break an existing lock
+    until the expected lifetime has expired.  Too long and other
+    processes will hang; too short and you'll end up trampling on
+    existing process locks -- and possibly corrupting data.  However
+    locks also support extending a lock's lifetime.  In a distributed
+    (NFS) environment, you also need to make sure that your clocks are
+    properly synchronized.
+
+    Each process laying claim to this resource lock will create their own
+    temporary lock file based on the path specified.  An optional lifetime
+    is the length of time that the process expects to hold the lock.
+
+    :param lockfile: The full path to the lock file.
+    :param lifetime: The expected maximum lifetime of the lock, as a
+        timedelta or integer number of seconds, relative to now.  Defaults
+        to 15 seconds.
+    :param separator: The separator character to use in the lock file's
+        file name.  Defaults to the vertical bar (`|`) on POSIX systems
+        and caret (`^`) on Windows.
         """
-        # This has to be defined before we call _set_claimfile().
+
+    def __init__(
+            self,
+            lockfile: str,
+            lifetime: Optional[Interval] = None,
+            separator: str = SEP
+            ):
+        """Create the resource lock using the given file name and lifetime."""
+        # The hostname has to be defined before we call _set_claimfile().
         self._hostname = socket.getfqdn()
         if lifetime is None:
             lifetime = DEFAULT_LOCK_LIFETIME
         self._lockfile = lockfile
-        self._lifetime = lifetime
+        # https://github.com/python/mypy/issues/3004
+        self.lifetime = lifetime                    # type: ignore
         # The separator must be set before we claim the lock.
         self._separator = separator
+        self._claimfile: str
         self._set_claimfile()
-        # For transferring ownership across a fork.
-        self._owned = True
         # For extending the set of NFS errnos that are retried in _read().
-        self._retry_errnos = []
+        self._retry_errnos: List[int] = []
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return '<%s %s [%s: %s] pid=%s at %#xx>' % (
             self.__class__.__name__,
             self._lockfile,
-            ('locked' if self.is_locked else 'unlocked'),
+            ('locked' if self._is_locked_no_refresh() else 'unlocked'),
             self._lifetime, os.getpid(), id(self))
 
     @property
-    def hostname(self):
+    def hostname(self) -> str:
         """The current machine's host name.
 
         :return: The current machine's hostname, as used in the `.details`
             property.
-        :rtype: str
         """
         return self._hostname
 
     @property
-    def details(self):
+    def details(self) -> Tuple[str, int, str]:
         """Details as read from the lock file.
 
-        :return: A 3-tuple of hostname, process id, file name.
-        :rtype: (str, int, str)
+        :return: A 3-tuple of hostname, process id, lock file name.
         :raises NotLockedError: if the lock is not acquired.
         """
         try:
             with open(self._lockfile) as fp:
                 filename = fp.read().strip()
-        except IOError as error:
+        except OSError as error:
             if error.errno in ERRORS:
-                raise NotLockedError('Details are unavailable')
+                raise NotLockedError('Details are unavailable') from error
             raise
         # Rearrange for signature.
         try:
             lockfile, hostname, pid, random_ignored = filename.split(
                 self._separator)
-        except ValueError:
-            raise NotLockedError('Details are unavailable')
+        except ValueError as error:
+            raise NotLockedError('Details are unavailable') from error
         return hostname, int(pid), lockfile
 
     @property
-    def lifetime(self):
+    def state(self) -> LockState:
+        """Infer the state of the lock."""
+        try:
+            with open(self._lockfile) as fp:
+                filename = fp.read().strip()
+        except FileNotFoundError:
+            return LockState.unlocked
+        try:
+            lockfile, hostname, pid_str, random_ignored = filename.split(
+                self._separator)
+            pid = int(pid_str)
+        except (ValueError, TypeError):
+            # The contents of the lock file is corrupt, so we can't know
+            # anything about the state of the lock.
+            return LockState.unknown
+        if hostname != self._hostname:
+            return LockState.unknown
+        if pid == os.getpid():
+            expired = self.expiration < datetime.now()
+            return LockState.ours_expired if expired else LockState.ours
+        if pid_exists(pid):
+            expired = self.expiration < datetime.now()
+            return LockState.theirs_expired if expired else LockState.unknown
+        return LockState.stale
+
+    @property
+    def lifetime(self) -> timedelta:
+        """The current lock life time."""
         return self._lifetime
 
     @lifetime.setter
-    def lifetime(self, lifetime):
-        self._lifetime = lifetime
+    def lifetime(self, lifetime: Interval) -> None:
+        if isinstance(lifetime, timedelta):
+            self._lifetime = lifetime
+        else:
+            self._lifetime = timedelta(seconds=lifetime)
 
-    def refresh(self, lifetime=None, unconditionally=False):
+    def refresh(self,
+                lifetime: Optional[Interval] = None,
+                *,
+                unconditionally: bool = False
+                ) -> None:
         """Refreshes the lifetime of a locked file.
 
         Use this if you realize that you need to keep a resource locked longer
         than you thought.
 
-        :param lifetime: If given, this sets the lock's new lifetime.  This
-            must be a datetime.timedelta.
-        :param unconditionally: When False (the default), a `NotLockedError`
+        :param lifetime: If given, this sets the lock's new lifetime.  It
+            represents the number of seconds into the future that the
+            lock's lifetime will expire, relative to now, or whenever it is
+            refreshed, either explicitly or implicitly.  If not given, the
+            original lifetime interval is used.
+        :param unconditionally: When False (the default), a ``NotLockedError``
             is raised if an unlocked lock is refreshed.
         :raises NotLockedError: if the lock is not set, unless optional
-            `unconditionally` flag is set to True.
+            ``unconditionally`` flag is set to True.
         """
         if lifetime is not None:
-            self._lifetime = lifetime
+            # https://github.com/python/mypy/issues/3004
+            self.lifetime = lifetime                # type: ignore
         # Do we have the lock?  As a side effect, this refreshes the lock!
         if not self.is_locked and not unconditionally:
             raise NotLockedError('{}: {}'.format(repr(self), self._read()))
 
-    def lock(self, timeout=None):
+    def lock(self, timeout: Optional[Interval] = None) -> None:
         """Acquire the lock.
 
         This blocks until the lock is acquired unless optional timeout is not
-        None, in which case a `TimeOutError` is raised when the timeout
+        None, in which case a ``TimeOutError`` is raised when the timeout
         expires without lock acquisition.
 
-        :param timeout: A datetime.timedelta indicating approximately how long
-            the lock acquisition attempt should be made.  None (the default)
-            means keep trying forever.
+        :param timeout: Approximately how long the lock acquisition attempt
+            should be made.  None (the default) means keep trying forever.
         :raises AlreadyLockedError: if the lock is already acquired.
-        :raises TimeOutError: if `timeout` is not None and the indicated time
-            interval expires without a lock acquisition.
+        :raises TimeOutError: if ``timeout`` is not None and the indicated
+            time interval expires without a lock acquisition.
         """
         if timeout is not None:
-            timeout_time = datetime.datetime.now() + timeout
+            if isinstance(timeout, int):
+                timeout = timedelta(seconds=timeout)
+            timeout_time = datetime.now() + timeout
         # Make sure the claim file exists, and that its contents are current.
         self._write()
         # XXX This next call can fail with an EPERM.  I have no idea why, but
-        # I'm nervous about wrapping this in a try/except.  It seems to be a
-        # very rare occurrence, only happens from cron, and has only(?) been
-        # observed on Solaris 2.6.
+        # I'm nervous about ignoring it.  It seems to be a very rare
+        # occurrence, only happens from cron, and has only(?) been observed on
+        # Solaris 2.6.
         self._touch()
         log.debug('laying claim: {}'.format(self._lockfile))
         # For quieting the logging output.
         loopcount = -1
         while True:
             loopcount += 1
-            # Create the hard link and test for exactly 2 links to the file.
+            # Create the hard link from the claim file to the lock file and
+            # test for a hard count of exactly 2 links.
             try:
                 os.link(self._claimfile, self._lockfile)
                 # If we got here, we know we got the lock, and never
@@ -253,18 +358,20 @@ class Lock:
                 pass
             # We did not acquire the lock, because someone else already has
             # it.  Have we timed out in our quest for the lock?
-            if timeout is not None and timeout_time < datetime.datetime.now():
+            if timeout is not None and timeout_time < datetime.now():
                 os.unlink(self._claimfile)
                 log.error('timed out')
                 raise TimeOutError('Could not acquire the lock')
             # Okay, we haven't timed out, but we didn't get the lock.  Let's
-            # find if the lock lifetime has expired.  Cache the release time
-            # to avoid race conditions.  (LP: #827052)
+            # find out if the lock lifetime has expired.  Cache the release
+            # time to avoid race conditions.  (LP: #827052)
             release_time = self._releasetime
-            if (release_time != -1 and
-                    datetime.datetime.now() > release_time + CLOCK_SLOP):
-                # Yes, so break the lock.
-                self._break()
+            if release_time != -1:
+                now = datetime.now()
+                future = cast(datetime, release_time) + CLOCK_SLOP
+                if now > future:
+                    # Yes, so break the lock.
+                    self._break()
                 log.error('lifetime has expired, breaking')
             # Okay, someone else has the lock, our claim hasn't timed out yet,
             # and the expected lock lifetime hasn't expired yet either.  So
@@ -273,14 +380,18 @@ class Lock:
                 log.debug('waiting for claim: {}'.format(self._lockfile))
             self._sleep()
 
-    def unlock(self, unconditionally=False):
+    # 2020-06-27(bwarsaw): `unconditionally` should really be a keyword-only
+    # argument, but those didn't exist when this library was originally
+    # written, and changing that now would be a needless backward incompatible
+    # API break.
+    def unlock(self, *, unconditionally: bool = False) -> None:
         """Release the lock.
 
-        :param unconditionally: When False (the default), a `NotLockedError`
+        :param unconditionally: When False (the default), a ``NotLockedError``
             is raised if this is called on an unlocked lock.
         :raises NotLockedError: if we don't own the lock, either because of
             unbalanced unlock calls, or because the lock was stolen out from
-            under us, unless optional `unconditionally` is True.
+            under us, unless optional ``unconditionally`` is True.
         """
         is_locked = self.is_locked
         if not is_locked and not unconditionally:
@@ -300,8 +411,15 @@ class Lock:
                 raise
         log.debug('unlocked: {}'.format(self._lockfile))
 
+    def _is_locked_no_refresh(self) -> bool:
+        """Don't refresh the lock, just return the status."""
+        # XXX Can the link count ever be > 2?
+        if self._linkcount != 2:
+            return False
+        return self._read() == self._claimfile
+
     @property
-    def is_locked(self):
+    def is_locked(self) -> bool:
         """True if we own the lock, False if we do not.
 
         Checking the status of the lock resets the lock's lifetime, which
@@ -315,91 +433,23 @@ class Lock:
             # how we can own the lock if we're not the owner.
             log.error('No permission to refresh the log')
             return False
-        # XXX Can the link count ever be > 2?
-        if self._linkcount != 2:
-            return False
-        return self._read() == self._claimfile
+        return self._is_locked_no_refresh()
 
-    def finalize(self):
-        """Unconditionally unlock the file."""
-        log.debug('finalize: {}'.format(self._lockfile))
-        self.unlock(unconditionally=True)
-
-    def __del__(self):
-        log.debug('__del__: {}'.format(self._lockfile))
-        if self._owned:
-            self.finalize()
-
-    def __enter__(self):
+    def __enter__(self) -> 'Lock':
         self.lock()
         return self
 
-    def __exit__(self, exc_type, exc_val, exc_tb):
+    def __exit__(self,
+                 exc_type: Optional[Type[BaseException]],
+                 exc_val: Optional[BaseException],
+                 exc_tb: Optional[TracebackType]
+                 ) -> Literal[False]:
         self.unlock()
         # Don't suppress any exception that might have occurred.
         return False
 
-    def transfer_to(self, pid):
-        """Transfer ownership of the lock to another process.
-
-        Use this only if you're transfering ownership to a child process across
-        a fork.  Use at your own risk, but it should be race-condition safe.
-        transfer_to() is called in the parent, passing in the pid of the child.
-        take_possession() is called in the child, and blocks until the parent
-        has transferred possession to the child.  disown() is used to set the
-        'owned' flag to False, and it is a disgusting wart necessary to make
-        forced lock acquisition work. :(
-
-        :param pid: The process id of the child process that will take
-            possession of the lock.
-        """
-        # First touch it so it won't get broken while we're fiddling about.
-        self._touch()
-        # Find out current claim's file name
-        winner = self._read()
-        # Now twiddle ours to the given pid.
-        self._set_claimfile(pid)
-        # Create a hard link from the global lock file to the claim file.
-        # This actually does things in reverse order of normal operation
-        # because we know that lockfile exists, and claimfile better not!
-        os.link(self._lockfile, self._claimfile)
-        # Now update the lock file to contain a reference to the new owner
-        self._write()
-        # Toggle off our ownership of the file so we don't try to finalize it
-        # in our __del__()
-        self._owned = False
-        # Unlink the old winner, completing the transfer.
-        os.unlink(winner)
-        # And do some sanity checks
-        assert self._linkcount == 2, (
-            'Unexpected link count: wanted 2, got {0:d}'.format(
-                self._linkcount))
-        assert self.is_locked, 'Expected to be locked'
-        log.debug('transferred the lock: {}'.format(self._lockfile))
-
-    def take_possession(self):
-        """Take possession of a lock from another process.
-
-        See `transfer_to()` for more information.
-        """
-        self._set_claimfile()
-        # Wait until the linkcount is 2, indicating the parent has completed
-        # the transfer.
-        while self._linkcount != 2 or self._read() != self._claimfile:
-            time.sleep(0.25)
-        log.debug('took possession of the lock: {}'.format(self._lockfile))
-
-    def disown(self):
-        """Disown this lock.
-
-        See `transfer_to()` for more information.
-        """
-        self._owned = False
-
-    def _set_claimfile(self, pid=None):
+    def _set_claimfile(self) -> None:
         """Set the _claimfile private variable."""
-        if pid is None:
-            pid = os.getpid()
         # Calculate a hard link file name that will be used to lay claim to
         # the lock.  We need to watch out for two Lock objects in the same
         # process pointing to the same lock file.  Without this, if you lock
@@ -407,31 +457,41 @@ class Lock:
         self._claimfile = self._separator.join((
             self._lockfile,
             self.hostname,
-            str(pid),
+            str(os.getpid()),
             str(random.randint(0, MAXINT)),
             ))
 
-    def _write(self):
+    def _write(self) -> None:
         """Write our claim file's name to the claim file."""
-        # Make sure it's group writable
+        # Make sure it's group writable.
         with open(self._claimfile, 'w') as fp:
             fp.write(self._claimfile)
 
     @property
-    def retry_errnos(self):
+    def retry_errnos(self) -> List[int]:
         """The set of errno values that cause a read retry."""
         return self._retry_errnos[:]
 
     @retry_errnos.setter
-    def retry_errnos(self, errnos):
+    def retry_errnos(self, errnos: List[int]) -> None:
         self._retry_errnos = []
         self._retry_errnos.extend(errnos)
 
     @retry_errnos.deleter
-    def retry_errnos(self):
+    def retry_errnos(self) -> None:
         self._retry_errnos = []
 
-    def _read(self):
+    @property
+    def expiration(self) -> datetime:
+        """The lock's expiration time, regardless of ownership."""
+        return datetime.fromtimestamp(os.stat(self._lockfile).st_mtime)
+
+    @property
+    def lockfile(self) -> str:
+        """Return the lock file name."""
+        return self._lockfile
+
+    def _read(self) -> Optional[str]:
         """Read the contents of our lock file.
 
         :return: The contents of the lock file or None if the lock file does
@@ -449,13 +509,13 @@ class Lock:
                 else:
                     return None
 
-    def _touch(self, filename=None):
+    def _touch(self, filename: Optional[str] = None) -> None:
         """Touch the claim file into the future.
 
         :param filename: If given, the file to touch, otherwise our claim file
             is touched.
         """
-        expiration_date = datetime.datetime.now() + self._lifetime
+        expiration_date = datetime.now() + self._lifetime
         t = time.mktime(expiration_date.timetuple())
         try:
             # XXX We probably don't need to modify atime, but this is easier.
@@ -465,22 +525,21 @@ class Lock:
                 raise
 
     @property
-    def _releasetime(self):
+    def _releasetime(self) -> Union[int, datetime]:
         """The time when the lock should be released.
 
         :return: The mtime of the file, which is when the lock should be
             released, or -1 if the lockfile doesn't exist.
         """
         try:
-            return datetime.datetime.fromtimestamp(
-                os.stat(self._lockfile).st_mtime)
+            return self.expiration
         except OSError as error:
             if error.errno in ERRORS:
                 return -1
             raise
 
     @property
-    def _linkcount(self):
+    def _linkcount(self) -> int:
         """The number of hard links to the lock file.
 
         :return: the number of hard links to the lock file, or -1 if the lock
@@ -493,7 +552,7 @@ class Lock:
                 return -1
             raise
 
-    def _break(self):
+    def _break(self) -> None:
         """Break the lock."""
         # First, touch the lock file.  This reduces but does not eliminate the
         # chance for a race condition during breaking.  Two processes could
@@ -505,17 +564,14 @@ class Lock:
         # Touching the lock could fail if the process breaking the lock and
         # the process that claimed the lock have different owners.  Don't do
         # that.
-        try:
+        with suppress(PermissionError):
             self._touch(self._lockfile)
-        except OSError as error:
-            if error.errno != errno.EPERM:
-                raise
-        # Get the name of the old winner's temp file.
+        # Get the name of the old winner's claim file.
         winner = self._read()
         # Remove the global lockfile, which actually breaks the lock.
         try:
             os.unlink(self._lockfile)
-        except OSError as error:
+        except OSError as error:                    # pragma: nocover
             if error.errno not in ERRORS:
                 raise
         # Try to remove the old winner's claim file, since we're assuming the
@@ -523,13 +579,13 @@ class Lock:
         # unlink their claim file -- this doesn't wreck the locking algorithm,
         # but will leave claim file turds laying around, a minor inconvenience.
         try:
-            if winner:
+            if winner:                              # pragma: nobranch
                 os.unlink(winner)
         except OSError as error:
-            if error.errno not in ERRORS:
+            if error.errno not in ERRORS:           # pragma: nocover
                 raise
 
-    def _sleep(self):
+    def _sleep(self) -> None:
         """Snooze for a random amount of time."""
         interval = random.random() * 2.0 + 0.01
         time.sleep(interval)
